# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def check_calibration_value_deviation(calibration_loop_vector, target_value, type_of_calibration,
                                      free_variable_calibration, path_of_working_directory_rce,
                                      current_workflow_name='UNICADOworkflow'):
    """ check_calibration_value_deviation checks the difference between current calibration value with given target value.

        The input list "calibration_loop_vector" contains all estimated ome or mtom values.

        The input value "target_value" contains the target value of current calibration type.
            * Contains a specified MTOM or OME target value to be hit

        The input string "type_of_calibration" contains the name of calibration type.
            * Contains the name OME or MTOM to check which outputs should be generated.

        The input int "free_variable_calibration" contains the status of free variable calibration.
            * If input integer is equal to 1 -> engine calibration is selected and is being performed

        The input string "path_of_working_directory_rce" contains the system path to rce working directory.

        The input string "current_workflow_name" contains the name of the current workflow execution.
            * If no input is given, default string 'UNICADOworkflow' is used.

        :param: calibration_loop_vector: input list
        :param: target_value: input value
        :param: type_of_calibration: input string
        :param: free_variable_calibration: input int
        :param: path_of_working_directory_rce: input string
        :param: current_workflow_name: input string
        :return: none
    """

    ''' imports for python '''
    import sys
    from datetime import datetime
    from sub_function.write_to_log_file import write_to_log_file

    ''' initialize local parameter '''
    log_file_list = []
    abort_flag = False
    max_deviation_calibration_value_to_target = 0.2

    # check if the current type of calibration is equal to 'OME'
    #  -> if true: -> check difference of ome calibration values
    if type_of_calibration == 'OME':
        if (abs(1 - calibration_loop_vector[-1]/target_value)) > max_deviation_calibration_value_to_target:
            print('  Target-OME: ' + str(target_value))
            print('  Current-OME: ' + str(calibration_loop_vector[-1]))
            print('  Deviation greater ' + str(max_deviation_calibration_value_to_target * 100) + ' percent!')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Target-OME: ' + str(target_value))
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Current-OME: ' + str(calibration_loop_vector[-1]))
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Deviation greater ' + str(max_deviation_calibration_value_to_target * 100)
                                 + ' percent!')

            # check if the last element of calibration value list is greater than given target value to hit
            if calibration_loop_vector[-1] > target_value:
                print('  Fuselage mass factor significantly too high!')
                log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                     'Fuselage mass factor significantly too high!')
            # else condition: -> the last element of calibration value list is lower or equal to the given target value
            else:
                print('  Fuselage mass factor significantly too low!')
                log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                     'Fuselage mass factor significantly too low!')

            print('  Select an appropriate start value! Program aborted')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Select an appropriate start value! Program aborted')
            abort_flag = True

    if type_of_calibration == 'MTOM':
        if (abs(1 - calibration_loop_vector[-1]/target_value)) > max_deviation_calibration_value_to_target:
            print('  Target-MTOM: ' + str(target_value))
            print('  Current-MTOM: ' + str(calibration_loop_vector[-1]))
            print('  Deviation greater ' + str(max_deviation_calibration_value_to_target * 100) + ' percent!')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Target-MTOM: ' + str(target_value))
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Current-MTOM: ' + str(calibration_loop_vector[-1]))
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Deviation greater ' + str(max_deviation_calibration_value_to_target * 100)
                                 + ' percent!')

            # check if the last element of calibration value list is greater than given target value to hit
            if calibration_loop_vector[-1] > target_value:
                if free_variable_calibration == 1:
                    print('  Engine efficiency factor significantly too high!')
                    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                         'Engine efficiency factor significantly too high!')
                else:
                    print('  Drag reduction significantly too low!')
                    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                         'Drag reduction significantly too low!')
            # else condition: -> the last element of calibration value list is lower or equal to the given target value
            else:
                if free_variable_calibration == 1:
                    print('  Engine efficiency factor significantly too low!')
                    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                         'Engine efficiency factor significantly too low!')
                else:
                    print('  Drag reduction significantly too high!')
                    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                         'Drag reduction significantly too high!')

            print('  Select an appropriate start value! Program aborted')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Select an appropriate start value! Program aborted')
            abort_flag = True

    # check if abort flag is equal to 'True' -> if true: -> write error messages to workflow log file
    if abort_flag:
        write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)
        sys.exit(1)

    return
