# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def damp_iteration(paths_and_names, mass_loop_vector, type_of_calibration):
    """ check_calibration_value_deviation checks the difference between current calibration value with given target value.

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input list "mass_loop_vector" contains all estimated mass values of current iteration loop.

        The input string "type_of_calibration" contains the name of calibration type.

        The input float "damped_value" contains the estimated damped iteration value.

        :param: paths_and_names: input list of lists
        :param: mass_loop_vector: input list
        :param: type_of_calibration: input string
        :return: damped_value: output float
    """

    ''' imports for python '''
    from datetime import datetime
    from inspect import currentframe, getframeinfo
    from sub_function.read_xml_file import read_xml_file
    from sub_function.write_to_log_file import write_to_log_file

    ''' read data for script execution '''
    aircraft_exchange_file = ([item for item in paths_and_names if 'aircraft_exchange_file' in item])[-1][-1]
    aircraft_project = ([item for item in paths_and_names if 'aircraft_project' in item])[-1][-1]
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]
    path_of_working_directory_rce = ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]

    ''' initialize local parameter '''
    log_file_list = []
    output_string = str()
    function_name = getframeinfo(currentframe()).function

    ''' estimate damped iteration value '''
    # the damped value is the average of the last two OME or MTOM values
    damped_value = 0.5 * (mass_loop_vector[-1] + mass_loop_vector[-2])

    ''' read data from aircraft exchange file '''
    frame_info = getframeinfo(currentframe())
    path = path_of_working_directory_rce + current_workflow_name + '/' + aircraft_project + '/' + aircraft_exchange_file
    root_of_aircraft_exchange_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                             aircraft_exchange_file, function_name, frame_info.lineno,
                                                             log_file_list, current_workflow_name)

    if type_of_calibration == 'OME':
        output_string = '  Use damped OME = ' + str(damped_value)
        # save damped ome to aircraft exchange file
        root_of_aircraft_exchange_tree.find(
            './analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass/value').text = str(damped_value)
        root_of_aircraft_exchange_tree.find('./analysis/masses_cg_inertia').set('tool_level', '1')

    if type_of_calibration == 'MTOM':
        output_string = '  Use damped MTOM = ' + str(damped_value)
        # save damped ome to aircraft exchange file
        root_of_aircraft_exchange_tree.find(
            './analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value').text = str(damped_value)
        root_of_aircraft_exchange_tree.find('./analysis/masses_cg_inertia').set('tool_level', '1')

    # write settings to configuration xml file
    xml_tree.write(path, encoding='utf-8')

    print(str(output_string) + ' kg for next iteration!')
    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': ' + output_string
                         + ' kg for next iteration!')
    log_file_list.append('')

    ''' write log-file to system '''
    write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)

    return damped_value
