# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def perform_ome_calibration(calibration_settings, parameter_for_design_case, ome_loop_vector, calibration_factors,
                            count_of_calibration, path_of_working_directory_rce,
                            current_workflow_name='UNIACDOworkflow'):
    """ perform_ome_calibration performs the ome target value calibration.

        The input list of lists "calibration_settings" contains the following values:
            *  [0][0] string: 'free_variable_OME_calibration',                   [0][1] int:   'switch for free variable OME calibration'
            *  [1][0] string: 'free_variable_MTOM_calibration',                  [1][1] int:   'switch for free variable MTOM calibration'
            *  [2][0] string: 'calibration_target_OME',                          [2][1] float: 'target value for OME calibration'
            *  [3][0] string: 'calibration_target_mtom',                         [3][1] float: 'target value for MTOM calibration'
            *  [4][0] string: 'reset_fuselage_mass_at_start',                    [4][1] int:   'switch to reset free OME calibration values at start'
            *  [5][0] string: 'reset_viscous_drag_counts_at_start',              [5][1] int:   'switch to reset aerodynamic MTOM calibration value at start'
            *  [6][0] string: 'reset_engine_fuel_flow_factor_at_start',          [6][1] int:   'switch to reset engine MTOM calibration value at start'
            *  [7][0] string: 'number_of_iterations_before_mtom_calibration',    [7][1] int:   'maximum number of OME calibration steps before performing a MTOM calibration'
            *  [8][0] string: 'max_iterations_before_exit_OME_calibration',      [8][1] int:   'maximum number of OME calibration steps before exit the calibration'
            *  [9][0] string: 'trunc_Error_factor',                              [9][1] int:   'value for termination criterion -> as factor regarding convergence criteria'
            * [10][0] string: 'damp_MTOM_calibration_lever',                    [10][1] int:   'switch for damping MTOM calibration lever arm'
            * [11][0] string: 'x_rel_max_for_MTOM_damp',                        [11][1] float: 'value of relative MTOM calibration lever arm'
            * [12][0] string: 'damp_OME_calibration_lever',                     [12][1] int:   'Switch for damping OME calibration lever arm'
            * [13][0] string: 'x_rel_max_for_OME_damp',                         [13][1] float: 'value of relative OME calibration lever arm'
            * [14][0] string: 'residual_iteration_value',                       [14][1] float: 'initial value of ome calibration residual'
            * [15][0] string: 'residual_iteration_value_to_target',             [15][1] float: 'initial value of ome calibration residual to target'
            * [16][0] string: 'calibration_factor_temp',                        [16][1] float: 'temporary value of ome calibration factor'
            * [17][0] string: 'calibration_factor_flag',                        [17][1] float: 'current value of ome calibration flag'
            * [18][0] string: 'calibration_factor_low'                          [18][1] float: 'value of lower calibration factor'
            * [19][0] string: 'calibration_factor_high'                         [19][1] float: 'value of higher calibration factor'
            * [20][0] string: 'calibration_factor_act',                         [20][1] float: 'value of actual calibration factor'
            * [21][0] string: 'calibration_factor_last',                        [21][1] float: 'value of last calibration factor'
            * [22][0] string: 'calibration_value_flag',                         [22][1] bool:  'current status of calibration value flag'
            * [23][0] string: 'calibration_value_low',                          [23][1] float: 'lower calibration value'
            * [24][0] string: 'calibration_value_high'                          [24][1] float: 'higher calibration value'
            * [25][0] string: 'calibration_value_act'                           [25][1] float: 'actual calibration value'
            * [26][0] string: 'calibration_value_last',                         [26][1] float: 'last calibration value'
            * [27][0] string: 'aux_interpolation_flag',                         [27][1] bool:  'status of auxiliary interpolation flag'
            * [28][0] string: 'calibration_value_over_target_exist',            [28][1] bool:  'status of calibration value over target value'
            * [29][0] string: 'calibration_value_under_target_exist',           [29][1] bool:  'status of calibration value under target value'
            * [30][0] string: 'calibration_value_over_target_update_exist',     [30][1] bool:  'status of calibration value update over target value'
            * [31][0] string: 'calibration_value_under_target_update_exist',    [31][1] bool:  'status of calibration value update under target value'
            * [32][0] string: 'number_reverse_effects',                         [32][1] int:   'number of reverse effects which are performed during calibration'
            * [33][0] string: 'number_of_final_runs',                           [33][1] int:   'number of final runs performed during calibration'
            * [34][0] string: 'x_relative'                                      [34][1] float: 'value of relative lever arm in x direction'
            * [35][0] string: 'max_delta_calibration_factor'                    [35][1] float: 'value of maximum allowed delta for calibration factor'
            * [36][0] string: 'calibrate_ome_flag',                             [36][1] bool:  'flag for ome calibration'
            * [37][0] string: 'calibrate_mtom_flag',                            [37][1] bool:  'flag for mtom calibration'
            * [38][0] string: 'final_fuselage_mass_factor',                     [38][1] float: 'value of final fuselage mass factor'
            * [39][0] string: 'current_value',                                  [39][1] float: 'value of current operating mass empty'
            * [40][0] string: 'last_value',                                     [40][1] float: 'value of last operating mass empty'
            * [41][0] string: 'iteration_flag'                                  [41][1] int:   'number of the status of calibration successful'

        The input list of lists "parameter_for_design_case" contains the following values:
            *  [0][0] string: 'program_mode',                     [0][1] string: 'Value of program execution mode - either standard_aircraft_design, or parameter_study, or optimization'
            *  [1][0] string: 'design_mode',                      [1][1] int:   'value to check which design mode is selected'
            *  [2][0] string: 'automatic_trim',                   [2][1] int:   'value to switch off and on the automatic trim function of the workflow'
            *  [3][0] string: 'convergence_criteria',             [3][1] float: 'value of convergence criteria of ome and mtom iteration'
            *  [4][0] string: 'number_of_max_iteration',          [4][1] int:   'value to set maximum number of design sizing iteration steps'
            *  [5][0] string: 'bridge_constraint_analyzer',       [5][1] int: 'value to switch off and on the execution of constraint_analysis module in workflow for each iteration step'
            *  [6][0] string: 'damp_mtom_iteration'               [6][1] bool:  'switch to activate damping of mtom iteration loop'
            *  [7][0] string: 'parallel_execution'                [7][1] int:   'switch to activate the parallel execution mode'
            *  [8][0] string: 'count_of_iteration',               [8][1] int:   'current number of iteration of design sizing loop'
            *  [9][0] string: 'count_of_mission_study_loop',      [9][1] int:   'current number of iteration of mission study loop'
            * [10][0] string: 'parameter_name',                  [10][1] string: 'name of the current input parameter if parameter study will be performed'
            * [11][0] string: 'global_loop_counter',             [11][1] int:    'global loop counter for parameter study mode'
            * [12][0] string: 'study_counter',                   [12][1] int:    'study counter of current input parameter if parameter study will be performed'
            * [13][0] string: 'new_value_of_input',              [13][1] float:  'value of current parameter study input parameter'

        The input list "ome_loop_vector" contains all estimated ome values.

        The input list "calibration_factors" contains all estimated calibration factors.

        The input value "count_of_calibration" contains the current number of calibration loop.

        The input string "path_of_working_directory" contains the system path to UNICADO installation directory.

        The input string "path_of_working_directory_rce" contains the system path to rce working directory.

        The input string "current_workflow_name" contains the name of the current workflow execution.
            * If no input is given, default string 'UNICADOworkflow' is used.

        The output list of lists "calibration_settings" contains the following values:
            *  [0][0] string: 'free_variable_OME_calibration',                   [0][1] int:   'switch for free variable OME calibration'
            *  [1][0] string: 'free_variable_MTOM_calibration',                  [1][1] int:   'switch for free variable MTOM calibration'
            *  [2][0] string: 'calibration_target_OME',                          [2][1] float: 'target value for OME calibration'
            *  [3][0] string: 'calibration_target_mtom',                         [3][1] float: 'target value for MTOM calibration'
            *  [4][0] string: 'reset_fuselage_mass_at_start',                    [4][1] int:   'switch to reset free OME calibration values at start'
            *  [5][0] string: 'reset_viscous_drag_counts_at_start',              [5][1] int:   'switch to reset aerodynamic MTOM calibration value at start'
            *  [6][0] string: 'reset_engine_fuel_flow_factor_at_start',          [6][1] int:   'switch to reset engine MTOM calibration value at start'
            *  [7][0] string: 'number_of_iterations_before_mtom_calibration',    [7][1] int:   'maximum number of OME calibration steps before performing a MTOM calibration'
            *  [8][0] string: 'max_iterations_before_exit_OME_calibration',      [8][1] int:   'maximum number of OME calibration steps before exit the calibration'
            *  [9][0] string: 'trunc_Error_factor',                              [9][1] int:   'value for termination criterion -> as factor regarding convergence criteria'
            * [10][0] string: 'damp_MTOM_calibration_lever',                    [10][1] int:   'switch for damping MTOM calibration lever arm'
            * [11][0] string: 'x_rel_max_for_MTOM_damp',                        [11][1] float: 'value of relative MTOM calibration lever arm'
            * [12][0] string: 'damp_OME_calibration_lever',                     [12][1] int:   'Switch for damping OME calibration lever arm'
            * [13][0] string: 'x_rel_max_for_OME_damp',                         [13][1] float: 'value of relative OME calibration lever arm'
            * [14][0] string: 'residual_iteration_value',                       [14][1] float: 'initial value of ome calibration residual'
            * [15][0] string: 'residual_iteration_value_to_target',             [15][1] float: 'initial value of ome calibration residual to target'
            * [16][0] string: 'calibration_factor_temp',                        [16][1] float: 'temporary value of ome calibration factor'
            * [17][0] string: 'calibration_factor_flag',                        [17][1] float: 'current value of ome calibration flag'
            * [18][0] string: 'calibration_factor_low'                          [18][1] float: 'value of lower calibration factor'
            * [19][0] string: 'calibration_factor_high'                         [19][1] float: 'value of higher calibration factor'
            * [20][0] string: 'calibration_factor_act',                         [20][1] float: 'value of actual calibration factor'
            * [21][0] string: 'calibration_factor_last',                        [21][1] float: 'value of last calibration factor'
            * [22][0] string: 'calibration_value_flag',                         [22][1] bool:  'current status of calibration value flag'
            * [23][0] string: 'calibration_value_low',                          [23][1] float: 'lower calibration value'
            * [24][0] string: 'calibration_value_high'                          [24][1] float: 'higher calibration value'
            * [25][0] string: 'calibration_value_act'                           [25][1] float: 'actual calibration value'
            * [26][0] string: 'calibration_value_last',                         [26][1] float: 'last calibration value'
            * [27][0] string: 'aux_interpolation_flag',                         [27][1] bool:  'status of auxiliary interpolation flag'
            * [28][0] string: 'calibration_value_over_target_exist',            [28][1] bool:  'status of calibration value over target value'
            * [29][0] string: 'calibration_value_under_target_exist',           [29][1] bool:  'status of calibration value under target value'
            * [30][0] string: 'calibration_value_over_target_update_exist',     [30][1] bool:  'status of calibration value update over target value'
            * [31][0] string: 'calibration_value_under_target_update_exist',    [31][1] bool:  'status of calibration value update under target value'
            * [32][0] string: 'number_reverse_effects',                         [32][1] int:   'number of reverse effects which are performed during calibration'
            * [33][0] string: 'number_of_final_runs',                           [33][1] int:   'number of final runs performed during calibration'
            * [34][0] string: 'x_relative'                                      [34][1] float: 'value of relative lever arm in x direction'
            * [35][0] string: 'max_delta_calibration_factor'                    [35][1] float: 'value of maximum allowed delta for calibration factor'
            * [36][0] string: 'calibrate_ome_flag',                             [36][1] bool:  'flag for ome calibration'
            * [37][0] string: 'calibrate_mtom_flag',                            [37][1] bool:  'flag for mtom calibration'
            * [38][0] string: 'final_fuselage_mass_factor',                     [38][1] float: 'value of final fuselage mass factor'
            * [39][0] string: 'current_value',                                  [39][1] float: 'value of current operating mass empty'
            * [40][0] string: 'last_value',                                     [40][1] float: 'value of last operating mass empty'
            * [41][0] string: 'iteration_flag'                                  [41][1] int:   'number of the status of calibration successful'

        The output list "ome_loop_vector" contains all estimated ome values.

        The output list "calibration_factors" contains all estimated calibration factors.

        The output bool "iteration_flag" contains the status of calibration.

        :param: calibration_settings: input list of lists
        :param: parameter_for_design_case: input list of lists
        :param: ome_loop_vector: input list
        :param: calibration_factors: input list
        :param: count_of_calibration: input value
        :param: path_of_working_directory: input string
        :param: path_of_working_directory_rce: input string
        :param: current_workflow_name: input string
        :return: calibration_settings: output list of lists, ome_loop_vector: output list, calibration_factors: output list, iteration_flag: output bool
    """

    ''' imports for python '''
    import sys
    from inspect import currentframe, getframeinfo
    from datetime import datetime
    from sub_function.read_xml_file import read_xml_file
    from sub_function.write_to_log_file import write_to_log_file
    from calibration.extrapolation_failed import extrapolation_failed
    from calibration.set_calibration_factor import set_calibration_factor
    from calibration.set_calibration_values import set_calibration_values
    from calibration.aux_interpolation_handling import aux_interpolation_handling
    from calibration.set_new_calibration_factor import set_new_calibration_factor
    from calibration.check_calibration_value_target import check_calibration_value_target
    from calibration.check_calibration_value_deviation import check_calibration_value_deviation

    ''' read data for script execution '''
    target_value = ([item for item in calibration_settings if 'calibration_target_OME' in item])[-1][-1]
    trunc_error_factor = ([item for item in calibration_settings if 'trunc_Error_factor' in item])[-1][-1]
    free_variable_ome_calibration =\
        int(([item for item in calibration_settings if 'free_variable_OME_calibration' in item])[-1][-1])

    convergence_criteria = ([item for item in parameter_for_design_case if 'convergence_criteria' in item])[-1][-1]

    ''' initialize local parameter '''
    log_file_list = []
    function_name = getframeinfo(currentframe()).function

    ''' check if deviation OME to target is too large '''
    check_calibration_value_deviation(ome_loop_vector, target_value, 'OME', free_variable_ome_calibration,
                                      path_of_working_directory_rce, current_workflow_name)

    ''' handling of auxiliary interpolation from previous run '''
    calibration_settings = aux_interpolation_handling(calibration_settings, ome_loop_vector, 'OME',
                                                      path_of_working_directory_rce, current_workflow_name)

    ''' set calibration values from previous run '''
    calibration_settings = set_calibration_values(calibration_settings, ome_loop_vector, 'OME', convergence_criteria,
                                                  count_of_calibration, path_of_working_directory_rce,
                                                  current_workflow_name)

    ''' set fuselage mass factors '''
    calibration_settings, calibration_factors = set_calibration_factor(calibration_settings, calibration_factors,
                                                                       count_of_calibration, 3)

    ''' check, whether OME is above or below OME_target '''
    calibration_settings, calibration_factors = check_calibration_value_target(calibration_settings,
                                                                               calibration_factors, 'OME', 0, 0,
                                                                               path_of_working_directory_rce,
                                                                               current_workflow_name)

    ''' set new fuselage mass factor '''
    calibration_settings, calibration_factors = set_new_calibration_factor(calibration_settings, calibration_factors,
                                                                           0.00, count_of_calibration,
                                                                           convergence_criteria, 'OME',
                                                                           path_of_working_directory_rce,
                                                                           current_workflow_name)

    ''' add current temporary calibration factor to calibration factor array '''
    calibration_factors.append(([item for item in calibration_settings if 'calibration_factor_temp' in item])[-1][-1])

    ''' reload calibration variables from calibration_settings list of list '''
    calibration_value_over_target_update_exist = \
        ([item for item in calibration_settings if 'calibration_value_over_target_update_exist' in item])[-1][-1]
    calibration_value_under_target_update_exist = \
        ([item for item in calibration_settings if 'calibration_value_under_target_update_exist' in item])[-1][-1]
    damp_ome_calibration_lever = \
        ([item for item in calibration_settings if 'damp_OME_calibration_lever' in item])[-1][-1]
    calibration_factor_temp = \
        ([item for item in calibration_settings if 'calibration_factor_temp' in item])[-1][-1]
    calibration_factor_low = \
        ([item for item in calibration_settings if 'calibration_factor_low' in item])[-1][-1]
    calibration_factor_high = \
        ([item for item in calibration_settings if 'calibration_factor_high' in item])[-1][-1]
    calibration_factor_act = \
        ([item for item in calibration_settings if 'calibration_factor_act' in item])[-1][-1]
    x_relative = float(([item for item in calibration_settings if 'x_relative' in item])[-1][-1])

    ''' catch special case, if extrapolation fails '''
    if (calibration_value_over_target_update_exist == 1 and calibration_value_under_target_update_exist == 1 and
            (calibration_factor_temp < calibration_factor_low or calibration_factor_temp > calibration_factor_high)):
        calibration_target_ome = ([item for item in calibration_settings if 'calibration_target_OME' in item])[-1][-1]
        # call function for output preparation if the extrapolation failed
        calibration_settings = extrapolation_failed(calibration_settings, calibration_target_ome,
                                                    path_of_working_directory_rce, current_workflow_name)

    ''' safety query check for too large change of the fuselage mass factors '''
    if count_of_calibration - 1 == 2:
        max_delta_calibration_factor = 0.3

    else:
        max_delta_calibration_factor = 0.08

    if abs(1 - calibration_factor_temp / calibration_factor_act) > max_delta_calibration_factor:
        print('  Change FuselageMassFactor (= ' + str(abs(1 - calibration_factor_temp / calibration_factor_act))
              + ') greater than permitted deviation (' + str(max_delta_calibration_factor) + ')!')
        log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                             'Change FuselageMassFactor (= '
                             + str(abs(1 - calibration_factor_temp / calibration_factor_act))
                             + ') greater than permitted deviation (' + str(max_delta_calibration_factor) + ')!')

        if x_relative > 1 and not damp_ome_calibration_lever == 1:
            print('  x_rel = ' + str(x_relative) + ' might be too high, try switching Damp_OME_Calibration_Lever on.')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   x_rel = ' +
                                 str(x_relative) + ' might be too high, try switching Damp_OME_Calibration_Lever on.')

        elif damp_ome_calibration_lever == 1:
            print('  Damp_OME_Calibration_Lever is on. x_rel is damped to ' + str(x_relative)
                  + '  but might still be too high. Try adjusting x_rel_max.')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                                 'Damp_OME_Calibration_Lever is on. x_rel is damped to ' + str(x_relative)
                                 + '  but might still be too high. Try adjusting x_rel_max.')

        print('  Check design, increase <NumberOfIterations>. Abort program.')
        log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   '
                             'Check design, increase <NumberOfIterations>. Abort program.')

        # write function output to workflow log file
        write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)
        # abort program
        sys.exit(1)

    ([item for item in calibration_settings if 'max_delta_calibration_factor' in item])[-1][-1] = \
        max_delta_calibration_factor

    ''' save the current fuselage mass factor '''
    residual_ome_to_target = \
        ([item for item in calibration_settings if 'residual_iteration_value_to_target' in item])[-1][-1]
    residual_fuselage_factor = abs(1 - (calibration_factors[-2]/calibration_factors[-1]))

    # check if convergence criteria reached -> if true: -> prepare final steps
    if ((residual_ome_to_target < convergence_criteria * trunc_error_factor) and
       (residual_fuselage_factor < convergence_criteria * trunc_error_factor)):
        ([item for item in calibration_settings if 'calibrate_ome_flag' in item])[-1][-1] = int(0)
        ([item for item in calibration_settings if 'final_fuselage_mass_factor' in item])[-1][-1] = \
            calibration_factors[-1]

        print('  !!!Calibration successful!!!" ')
        log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ':   !!!Calibration successful!!!')
        # write function output to workflow log file
        write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)

    ''' write temporary fuselage mass factor to configuration file of mass estimation module '''
    # try to open weight_and_balance_analysis module configuration file
    frame_info = getframeinfo(currentframe())
    path = path_of_working_directory_rce + current_workflow_name + '/fuselage_design_conf.xml'
    root_of_mass_estimation_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce, 'fuselage_design',
                                                           function_name, frame_info.lineno, log_file_list,
                                                           current_workflow_name)

    # reset fuselage scaling factor in weight_and_balance_analysis configuration file to 1
    root_of_mass_estimation_tree.find(
        './program_settings/configuration/fidelity/fuselage_design_tu_berlin/general/mass_technology_factors/'
        'fuselage_structural_mass_technology_factor/value').text = str(calibration_factor_temp)

    # write settings to configuration xml file
    xml_tree.write(path, encoding='utf-8')

    return calibration_settings, ome_loop_vector, calibration_factors
