# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def save_calibration_results_to_file(paths_and_names, count_of_calibration, residual_to_target, type_of_calibration):
    """ check_calibration_value_deviation checks the difference between current calibration value with given target value.

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input int "count_of_calibration" contains the current number of calibration loop.

        The input float "residual_to_target" contains the current residual to calibration target.

        The input string "type_of_calibration" contains the name of calibration type.

        :param: paths_and_names: input list of lists
        :param: count_of_calibration: input int
        :param: residual_to_target: input float
        :param: type_of_calibration: input string
        :return: none
    """

    ''' imports for python '''
    import os
    from inspect import currentframe, getframeinfo
    from sub_function.read_xml_file import read_xml_file
    from sub_function.get_mission_energy_specific_segment import get_mission_energy_specific_segment

    ''' read data for script execution '''
    aircraft_exchange_file = ([item for item in paths_and_names if 'aircraft_exchange_file' in item])[-1][-1]
    aircraft_project = ([item for item in paths_and_names if 'aircraft_project' in item])[-1][-1]
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]
    path_of_working_directory_rce =\
        ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]

    ''' initialize local parameter '''
    results = []
    log_file_list = []
    function_name = getframeinfo(currentframe()).function
    reporting_of_ome_calibration = path_of_working_directory_rce + current_workflow_name\
        + '/temporaryResults/reporting/'

    ''' open and read aircraft exchange file '''
    frame_info = getframeinfo(currentframe())
    path = path_of_working_directory_rce + current_workflow_name + '/' + aircraft_project + '/' + aircraft_exchange_file
    root_of_aircraft_exchange_tree, _ = read_xml_file(path, path_of_working_directory_rce,
                                                             aircraft_exchange_file, function_name, frame_info.lineno,
                                                             log_file_list, current_workflow_name)

    if type_of_calibration == 'OME':
        m_take_off_max = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value")
        if m_take_off_max is not None:
            m_take_off_max = float(m_take_off_max.text)
        else:
            m_take_off_max = 0
        
        m_operating_empty = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass/value")
        if m_operating_empty is not None:
            m_operating_empty = float(m_operating_empty.text)
        else:
            m_operating_empty = 0

        mission_energy = get_mission_energy_specific_segment(
            root_of_aircraft_exchange_tree, 'design_mission', 'loaded_mission_energy/mission_energy')
        
        x_center_of_gravity = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/center_of_gravity/x/value")
        if x_center_of_gravity is not None:
            x_center_of_gravity = float(x_center_of_gravity.text)
        else:
            x_center_of_gravity = 0

        if count_of_calibration == 0:
            if not os.path.isdir(reporting_of_ome_calibration):
                os.makedirs(reporting_of_ome_calibration)

            results.append('count_of_iteration; '
                           'm_take_off_max in kg; '
                           'residual_m_take_off_max; '
                           'm_operating_empty in kg; '
                           'residual_m_operating_empty; '
                           'mission_energy in J; '
                           'residual_mission_energy; '
                           'x_center_of_gravity in m; '
                           'residual_x_center_of_gravity;')

        results.append(str(count_of_calibration) + ';' +
                       str(round(m_take_off_max, 2)) + ';' +
                       'nan;' +
                       str(round(m_operating_empty, 2)) + ';' +
                       str(round(residual_to_target, 6)) + ';' +
                       str(round(mission_energy, 2)) + ';' +
                       'nan;' +
                       str(round(x_center_of_gravity, 3)) + ';' +
                       'nan;')

    results_file = open(reporting_of_ome_calibration + 'OME_calibration_results.csv', 'a')
    for row in results:
        results_file.write(row + '\n')
    results_file.close()

    return
