# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def initialize_design_sizing_loop(paths_and_names, parameter_for_design_case, control_settings):
    """ initialize_design_sizing_loop initialize design sizing loop and auto-trim function.
        Set initial values of sizing loop and auto-trim parameter

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input list of lists "parameter_for_design_case" contains the following values:
            *  [0][0] string: 'program_mode',                     [0][1] string: 'Value of program execution mode - either standard_aircraft_design, or parameter_study, or optimization'
            *  [1][0] string: 'design_mode',                      [1][1] int:   'value to check which design mode is selected'
            *  [2][0] string: 'automatic_trim',                   [2][1] int:   'value to switch off and on the automatic trim function of the workflow'
            *  [3][0] string: 'convergence_criteria',             [3][1] float: 'value of convergence criteria of ome and mtom iteration'
            *  [4][0] string: 'number_of_max_iteration',          [4][1] int:   'value to set maximum number of design sizing iteration steps'
            *  [5][0] string: 'bridge_constraint_analyzer',       [5][1] int: 'value to switch off and on the execution of constraint_analysis module in workflow for each iteration step'
            *  [6][0] string: 'damp_mtom_iteration'               [6][1] bool:  'switch to activate damping of mtom iteration loop'
            *  [7][0] string: 'parallel_execution'                [7][1] int:   'switch to activate the parallel execution mode'
            *  [8][0] string: 'count_of_iteration',               [8][1] int:   'current number of iteration of design sizing loop'
            *  [9][0] string: 'count_of_mission_study_loop',      [9][1] int:   'current number of iteration of mission study loop'
            * [10][0] string: 'parameter_name',                  [10][1] string: 'name of the current input parameter if parameter study will be performed'
            * [11][0] string: 'global_loop_counter',             [11][1] int:    'global loop counter for parameter study mode'
            * [12][0] string: 'study_counter',                   [12][1] int:    'study counter of current input parameter if parameter study will be performed'
            * [13][0] string: 'new_value_of_input',              [13][1] float:  'value of current parameter study input parameter'

        The input list of lists "control_settings" contains the following values:
            *  [0][0] string: 'delete_plots_before_run',                              [0][1] int: 'switch to delete old plots before running'
            *  [1][0] string: 'delete_engines_before_run',                            [1][1] int: 'switch to delete old engine data before running'
            *  [2][0] string: 'clear_old_log_files_at_start',                         [2][1] int: 'switch to delete old log-files before running'
            *  [3][0] string: 'plot_output_on',                                       [3][1] int: 'switch to enable or disable plot generation'
            *  [4][0] string: 'switch_off_plots',                                     [4][1] int: 'switch to enable generation of plots of each iteration step'
            *  [5][0] string: 'report_output_on'                                      [5][1] int: 'switch to enable or disable report generation'
            *  [6][0] string: 'switch_off_reports',                                   [6][1] int: 'switch to enable generation of reports of each iteration step'
            *  [7][0] string: 'tex_report_on'                                         [7][1] int: 'switch to enable or disable tex-report generation'
            *  [8][0] string: 'switch_off_tex',                                       [8][1] int: 'switch to enable generation of tex-reports of each iteration step'
            *  [9][0] string: 'status_flag_to_skip_design_sizing'                     [9][1] int: 'status flag for plot generation of design sizing loop - iniitalized with 0'
            * [10][0] string: 'status_flag_to_skip_study_mission',                   [10][1] int: 'status flag for plot generation of study mission loop - iniitalized with 0'
            * [11][0] string: 'check_settings_for_design_logic',                     [11][1] int: 'value to switch off and on the checks of the design logic'
            * [12][0] string: 'use_control_settings_for_sub_programs',               [12][1] int: 'value for switching off and on the override control settings of subroutines'
            * [13][0] string: 'use_range_type_specific_factors',                     [13][1] int: 'value for switch of range scenario to set range dependent parameters'
            * [14][0] string: 'use_configs_from_existing_project',                   [14][1] int: 'value to activate copy of configuration files from existing project before running'
            * [15][0] string: 'use_engine_from_existing_project',                    [15][1] int: 'value to activate copy of engine data from existing project before running'
            * [16][0] string: 'save_configs_to_result_folder',                       [16][1] int: 'value to switch off and on the saving of configuration files in the result folder'
            * [17][0] string: 'save_log_files_to_result_folder',                     [17][1] int: 'value to switch off and on the saving of log files in the result folder'
            * [18][0] string: 'save_acft_xml_to_result_folder',                      [18][1] int: 'value to switch off and on the saving of aircraft exchange file in the result folder'
            * [19][0] string: 'save_acft_xml_after_each_iteration',                  [19][1] int: 'value to switch off and on the saving of aircraft exchange file in the result folder after each step of iteration'
            * [20][0] string: 'write_info_files',                                    [20][1] int: 'switch to activate the writing of info files after each tool execution'
            * [21][0] string: 'write_info_files_only_at_last_step_of_iteration',     [21][1] int: 'switch to activate the writing of info files only after last step of sizing loop'
            * [22][0] string: 'program_mode',                                        [22][1] string: 'Value of program execution mode - either standard_aircraft_design, or parameter_study, or optimization'

        The output list of lists "design_variables" contains the following values:
            * [0][0] string: 'm_take_off_max',          [0][1] float: 'value of initial maximum take off mass'
            * [1][0] string: 'm_operating_empty',       [1][1] float: 'value of initial operating empty mass'
            * [2][0] string: 'mission_energy',          [2][1] float: 'value of initial consumed mission energy of design mission'
            * [3][0] string: 'x_center_of_gravity',     [3][1] float: 'value of initial x-position of center of gravity'

        The output list of lists "trim_parameter_list" contains the following values:
            * [0][0] string: 'i_stab',                          [0][1] array: 'array of values of angle of incidence of horizontal stabilizer'
            * [1][0] string: 'i_stab_tmp',                      [1][1] float: 'value of temporary angle of incidence of horizontal stabilizer'
            * [2][0] string: 'i_stab_last',                     [2][1] float: 'value of last angle of incidence of horizontal stabilizer'
            * [3][0] string: 'cm_cruise',                       [3][1] array: 'array of values of moment coefficient of cruise flight'
            * [4][0] string: 'cm_cruise_current',               [4][1] float: 'value of temporary moment coefficient of cruise flight'
            * [5][0] string: 'cm_cruise_last',                  [5][1] float: 'value of last moment coefficient of cruise flight'
            * [6][0] string: 'flag_trim_successful',            [6][1] int:   'value to signalize that trim is successful'
            * [7][0] string: 'retries',                         [7][1] int:   'value to count the retries of trim function'
            * [8][0] String: 'trim_accuracy'                    [8][1] float: 'value of trim accuracy if the ohn convergence criterion should be used'

        The output list "mtom_loop_vector" contains all estimated mtom values.

        :param paths_and_names: input list of lists
        :param parameter_for_design_case: input list of lists
        :param control_settings: input list of lists
        :return design_variables: output list of lists, trim_parameter_list: output list of lists, mtom_loop_vector: output list
   """

    ''' imports for python '''
    import os
    import shutil
    from datetime import datetime
    from inspect import currentframe, getframeinfo
    from sub_function.read_xml_file import read_xml_file
    from sub_function.get_mission_energy_specific_segment import get_mission_energy_specific_segment
    from sub_function.write_to_log_file import write_to_log_file

    ''' create list for log-file string '''
    log_file_list = ['',
                     '****************************************** initialize design sizing loop '
                     '******************************************']

    ''' read data for script execution '''
    aircraft_exchange_file = ([item for item in paths_and_names if 'aircraft_exchange_file' in item])[-1][-1]
    aircraft_project = ([item for item in paths_and_names if 'aircraft_project' in item])[-1][-1]
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]
    path_of_working_directory = ([item for item in paths_and_names if 'path_of_working_directory' in item])[-1][-1]
    path_of_working_directory_rce = \
        ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]

    automatic_trim = ([item for item in parameter_for_design_case if 'automatic_trim' in item])[-1][-1]
    design_mode = int(([item for item in parameter_for_design_case if 'design_mode' in item])[-1][-1])
    number_of_max_iteration =\
        ([item for item in parameter_for_design_case if 'number_of_max_iteration' in item])[-1][-1]

    switch_off_plots = ([item for item in control_settings if 'switch_off_plots' in item])[-1][-1]
    switch_off_reports = ([item for item in control_settings if 'switch_off_reports' in item])[-1][-1]
    switch_off_tex = ([item for item in control_settings if 'switch_off_tex' in item])[-1][-1]

    ''' initialize local parameter '''
    function_name = getframeinfo(currentframe()).function
    mtom_loop_vector = []

    print(os.path.isfile(path_of_working_directory_rce + current_workflow_name
                         + '/temp/workflowExecutionError.dat'))
    # check if design_mode not equal to 3 and no error in tool execution before
    # -> if true: -> design sizing loop should be executed -> prepare initial parameter for design sizing
    if not design_mode == 3 and not os.path.isfile(path_of_working_directory_rce + current_workflow_name
                                                   + '/temp/workflowExecutionError.dat'):
        ''' create list of files in working directory '''
        files_in_working_directory = os.listdir(path_of_working_directory)
        if switch_off_plots == 1 or switch_off_reports == 1 or switch_off_tex == 1:
            # loop for all elements in list of files
            for file_name in files_in_working_directory:
                if not (file_name == 'workingDirectoryRCE') and not (file_name == 'workflowResults') \
                        and not (file_name == 'projects') and not (file_name == 'cpacs_interface') and not (file_name == 'report_generator'):
                    # check if current list element is a directory
                    if os.path.isdir(path_of_working_directory + file_name):
                        # check whether a configuration xml file exists within the current list element
                        if os.path.isfile(path_of_working_directory_rce + current_workflow_name + '/'
                                          + file_name + '_conf.xml'):
                            # disable plot and report generation for sizing loop
                            frame_info = getframeinfo(currentframe())
                            path = path_of_working_directory_rce + current_workflow_name + '/' + file_name + '_conf.xml'
                            file_name_tree_root, xml_tree = read_xml_file(path, path_of_working_directory_rce, file_name,
                                                                          function_name,  frame_info.lineno,
                                                                          log_file_list, current_workflow_name)

                            if switch_off_plots == 1:
                                file_name_tree_root.find('./control_settings/plot_output/enable/value').text = \
                                    str(bool(0)).lower()
                            if switch_off_reports == 1:
                                file_name_tree_root.find('./control_settings/report_output/value').text = \
                                    str(bool(0)).lower()
                            if switch_off_tex == 1:
                                file_name_tree_root.find('./control_settings/tex_report/value').text = \
                                    str(bool(0)).lower()

                            # write settings to configuration xml file
                            xml_tree.write(path, encoding='utf-8')

        ''' open and read aircraft exchange file '''
        frame_info = getframeinfo(currentframe())
        path = path_of_working_directory_rce + current_workflow_name + '/' + aircraft_project \
            + '/' + aircraft_exchange_file
        root_of_aircraft_exchange_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                                 aircraft_exchange_file, function_name,
                                                                 frame_info.lineno, log_file_list,
                                                                 current_workflow_name)

        # set values to output variables
        m_take_off_max = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value")
        if m_take_off_max is not None:
            m_take_off_max = float(m_take_off_max.text)
        else:
            m_take_off_max = 0

        m_operating_empty = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass/value")
        if m_operating_empty is not None:
            m_operating_empty = float(m_operating_empty.text)
        else:
            m_operating_empty = 0

        mission_energy_mission_design = get_mission_energy_specific_segment(
            root_of_aircraft_exchange_tree, "design_mission", "loaded_mission_energy/mission_energy")

        x_center_of_gravity = root_of_aircraft_exchange_tree.find(
            "./analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x/value")
        if x_center_of_gravity is not None:
            x_center_of_gravity = float(x_center_of_gravity.text)
        else:
            x_center_of_gravity = 0

        design_variables = [['m_take_off_max', m_take_off_max],
                            ['m_operating_empty', m_operating_empty],
                            ['mission_energy', mission_energy_mission_design],
                            ['x_center_of_gravity', x_center_of_gravity]]

        # set current mtom value tp loop vector
        mtom_loop_vector.append(m_take_off_max)

        ''' check if auto-trim is activated '''
        # initialize trim parameter
        cm_cruise = []
        cm_cruise_last = float(0.0)
        cm_cruise_current = float(0.0)
        i_stab = []
        i_stab_tmp = float(0.0)
        flag_trim_successful = 0
        retries = int(0)
        trim_accuracy = float()

        ''' Check if auto trim is activated '''
        if automatic_trim:
            # initialize automated trimming
            print('Auto-Trim is activated!')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': Auto-Trim is activated!')
            number_of_lifting_surfaces = len(root_of_aircraft_exchange_tree.findall(
                "./component_design/empennage/specific/geometry/aerodynamic_surface"))
            name_of_lifting_surfaces = root_of_aircraft_exchange_tree.findall(
                './component_design/empennage/specific/geometry/aerodynamic_surface/name/value')

            # open and read workflow configuration file
            frame_info = getframeinfo(currentframe())
            path = path_of_working_directory_rce + current_workflow_name + '/unicado_workflow_conf.xml'
            root_of_workflow_tree, _ = read_xml_file(path, path_of_working_directory_rce, 'unicado_workflow',
                                                     function_name, frame_info.lineno, log_file_list,
                                                     current_workflow_name)

            trim_accuracy = float(1e-5)

            # check if number of lifting surfaces is greater than zero
            if number_of_lifting_surfaces > 0:
                # read names of lifting surfaces
                lifting_surface_list = []
                for liftingSurface in name_of_lifting_surfaces:
                    lifting_surface_list.append(liftingSurface.text)
                # check if only one horizontal stabiliser exists
                if (lifting_surface_list.count('horizontal_stabiliser') == 1):
                    automatic_trim = 1
                else:
                    print('Unknown designation for wing-tail arrangement \n' '(Name Wing: ' + lifting_surface_list[0]
                          + ', Name Stabiliser: ' + lifting_surface_list[1] + ').\n' 'Switch automatic trim off!')
                    log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': '
                                         'Unknown designation for wing-tail arrangement \n' '(Name Wing: '
                                         + lifting_surface_list[0] + ', Name Stabiliser: ' + lifting_surface_list[1]
                                         + ').\n' 'Switch automatic trim off!')
                    automatic_trim = 0

            # otherwise: check if module empennage_design is in working directory
            elif 'empennage_design' in files_in_working_directory:
                automatic_trim = 1

            # otherwise: check only one lifting surface is available -> auto-trim disabled
            else:
                print('Only one lifting surface available and empennage_design not activated. '
                      'This is not enough for an automated trimming. Switch off trimming.')
                log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': '
                                     'Only one lifting surface available and empennage_design not activated. '
                                     'This is not enough for an automated trimming. Switch off trimming.')
                automatic_trim = 0
        else:
            print('Auto-Trim is disabled!')
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': Auto-Trim is disabled!')

        ''' set start values '''
        i_stab.append(i_stab_tmp)
        i_stab_last = i_stab_tmp
        cm_cruise.append(i_stab_tmp)

        '''  create table for for trim parameter '''
        ([item for item in parameter_for_design_case if 'automatic_trim' in item])[-1][-1] = automatic_trim
        trim_parameter_list = [['i_stab', i_stab],
                               ['i_stab_tmp', float(i_stab_tmp)],
                               ['i_stab_last', float(i_stab_last)],
                               ['cm_cruise', cm_cruise],
                               ['cm_cruise_current', float(cm_cruise_current)],
                               ['cm_cruise_last', float(cm_cruise_last)],
                               ['flag_trim_successful', int(flag_trim_successful)],
                               ['retries', int(retries)],
                               ['trim_accuracy', trim_accuracy]]

        ''' crate temporary folder for aircraft exchange files of each sizing loop '''
        aircraft_exchange_files_of_sizing_loop = path_of_working_directory_rce + current_workflow_name \
            + '/temporaryResults/aircraft_exchange_files/aircraft_exchange_files_of_sizing_loop/'
        if not os.path.isdir(aircraft_exchange_files_of_sizing_loop):
            os.makedirs(aircraft_exchange_files_of_sizing_loop)

        ''' save clean aircraft exchange file '''
        shutil.copy(path_of_working_directory_rce + current_workflow_name + '/' + aircraft_project + '/'
                    + aircraft_exchange_file,
                    aircraft_exchange_files_of_sizing_loop + aircraft_exchange_file[:-4] + '_it0.xml')

        ''' start of iteration '''
        print('----- Start Iteration: Maximum steps of iteration to perform = '
              + str(number_of_max_iteration) + '! -----')

        ''' write log-file to system '''
        log_file_list.append('**************************************** end initialize design sizing loop '
                             '****************************************')
        log_file_list.append('')
        log_file_list.append('************************** Start Iteration: Maximum steps of iteration to perform = '
                             + str(number_of_max_iteration) + '! ***************************')
        log_file_list.append('')

        write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)

    # Else condition: design_mode is equal to 3 -> skip design sizing loop and perform mission study loop
    #  -> initialize design variables and trim parameter with zero
    else:
        design_variables = [['m_take_off_max', 0],
                            ['m_operating_empty', 0],
                            ['mission_energy', 0],
                            ['x_center_of_gravity', 0]]

        ([item for item in parameter_for_design_case if 'automatic_trim' in item])[-1][-1] = 0
        cm_cruise = [0]
        i_stab = [0]
        mtom_loop_vector = [0]
        trim_parameter_list = [['i_stab', i_stab],
                               ['i_stab_tmp', 0],
                               ['i_stab_last', 0],
                               ['cm_cruise', cm_cruise],
                               ['cm_cruise_current', 0],
                               ['cm_cruise_last', 0],
                               ['flag_trim_successful', 0],
                               ['retries', 0],
                               ['trim_accuracy', 0]]

    return design_variables, trim_parameter_list, mtom_loop_vector
