# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def save_design_sizing_results_to_file(paths_and_names, design_variables, count_of_iteration):
    """ save_design_sizing_results_to_file writes all design sizing results to output csv-file.

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input list of lists "design_variables" contains the following values:
            *  [0][0] string: 'm_take_off_max_last',                     [0][1] float: 'value of maximum take off mass of last iteration step'
            *  [1][0] string: 'm_operating_empty_last',                  [1][1] float: 'value of operating empty mass of last iteration step'
            *  [2][0] string: 'mission_energy_last',                     [2][1] float: 'value of mission energy of last iteration step'
            *  [3][0] string: 'x_center_of_gravity_last',                [3][1] float: 'value of x position of center of gravity of last iteration step'
            *  [4][0] string: 'cm_cruise',                               [4][1] float: 'value of the current moment coefficient'
            *  [5][0] string: 'i_stab',                                  [5][1] float: 'value of the current trimming angle of horizontal stabiliser'
            *  [6][0] string: 'm_take_off_max',                          [6][1] float: 'value of maximum take off mass of current iteration step'
            *  [7][0] string: 'm_operating_empty',                       [7][1] float: 'value of operating empty mass of current iteration step'
            *  [8][0] string: 'mission_energy',                          [8][1] float: 'value of mission energy of current iteration step'
            *  [9][0] string: 'x_center_of_gravity',                     [9][1] float: 'value of x position of center of gravity of current iteration step'
            * [10][0] string: 'residual_m_take_off_max',                [10][1] float: 'value of residual of maximum take off mass of current iteration step'
            * [11][0] string: 'residual_m_operating_empty',             [11][1] float: 'value of residual of operating empty mass of current iteration step'
            * [12][0] string: 'residual_mission_energy',                [12][1] float: 'value of residual of mission energy of current iteration step'
            * [13][0] string: 'residual_x_center_of_gravity',           [13][1] float: 'value of residual of x position of center of gravity of current iteration step'
            * [14][0] string: 'status_residual_m_take_off_max',         [14][1] bool:  'status of convergence for maximum take of mass'
            * [15][0] string: 'status_residual_m_operating_empty',      [15][1] bool:  'status of convergence for mass operating empty'
            * [16][0] string: 'status_residual_mission_energy',         [16][1] bool:  'status of convergence for mission energy'
            * [17][0] string: 'status_residual_x_center_of_gravity',    [17][1] bool:  'status of convergence for the aft x-position of the center of gravity'

        The input int "count_of_iteration" contains the current number of calibration loop.

        :param: paths_and_names: input list of lists
        :param: design_variables: input list of lists
        :param: count_of_iteration: input int
        :return: none
    """

    ''' imports for python '''
    import os

    ''' read data for script execution '''
    path_of_working_directory_rce = \
        ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]

    ''' initialize local parameter '''
    results = []
    reporting_of_ome_calibration = path_of_working_directory_rce + current_workflow_name \
        + '/temporaryResults/reporting/'

    ''' check if output temporary reporting folder exist '''
    if not os.path.isdir(reporting_of_ome_calibration):
        os.makedirs(reporting_of_ome_calibration)

    ''' prepare output data for csv-file '''
    # check if first loop of iteration -> if true: -> generate file header
    if count_of_iteration == 1:
        # generate header
        results.append('count_of_iteration; '
                       'm_take_off_max in kg; '
                       'residual_m_take_off_max; '
                       'm_operating_empty in kg; '
                       'residual_m_operating_empty; '
                       'mission_energy in J; '
                       'residual_mission_energy; '
                       'x_center_of_gravity in m; '
                       'residual_x_center_of_gravity; '
                       'cm_cruise; '
                       'i_stab in deg')

        # save initial values to csv-file
        results.append(str(count_of_iteration - 1) + ';' +
                       str(round(([item for item in design_variables if 'm_take_off_max_last' in item])[-1][-1], 2))
                       + ';' +
                       'nan;' +
                       str(round(([item for item in design_variables if 'm_operating_empty_last' in item])[-1][-1], 2))
                       + ';' +
                       'nan;' +
                       str(0.00) + ';' +
                       'nan;' +
                       str(0.000) + ';' +
                       'nan;' +
                       str(0.00000) + ';' +
                       str(0.0000) + ';')

    results.append(str(count_of_iteration) + ';' +
                   str(round(([item for item in design_variables if 'm_take_off_max' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in design_variables if 'residual_m_take_off_max' in item])[-1][-1], 6))
                   + ';' +
                   str(round(([item for item in design_variables if 'm_operating_empty' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in design_variables if 'residual_m_operating_empty' in item])[-1][-1], 6))
                   + ';' +
                   str(round(([item for item in design_variables if 'mission_energy' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in design_variables if 'residual_mission_energy' in item])[-1][-1], 6))
                   + ';' +
                   str(round(([item for item in design_variables if 'x_center_of_gravity' in item])[-1][-1], 3)) + ';' +
                   str(round(([item for item in design_variables if 'residual_x_center_of_gravity' in item])
                             [-1][-1], 6)) + ';' +
                   str(round(([item for item in design_variables if 'cm_cruise' in item])[-1][-1], 5)) + ';' +
                   str(round(([item for item in design_variables if 'i_stab' in item])[-1][-1], 4)))

    results_file = open(reporting_of_ome_calibration + 'sizing_loop_results.csv', 'a')
    for row in results:
        results_file.write(row + '\n')
    results_file.close()

    return
