# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def save_mission_study_results_to_file(paths_and_names, mission_study_variables, count_of_mission_study_loop):
    """ save_mission_study_results_to_file writes all mission study results to output csv-file.

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input list of lists "mission_study_variables" contains the following values:
            *  [0][0] string: 'm_take_off_max_study',                    [0][1] float: 'value of maximum take off mass of current mission study loop'
            *  [1][0] string: 'mission_energy_study',                    [1][1] float: 'value of mission energy of current mission study loop'
            *  [2][0] string: 'range_study',                             [2][1] float: 'value of range of current mission study loop'
            *  [3][0] string: 'fl_initial_cruise_study',                 [3][1] float: 'value of cruise flight level of current mission study loop'
            *  [4][0] string: 'residual_m_take_off_max_study',           [4][1] float: 'value of residual of maximum take off mass of current mission study loop'
            *  [5][0] string: 'residual_mission_energy_study',           [5][1] float: 'value of residual of mission energy of current mission study loop'
            *  [6][0] string: 'residual_range_study',                    [6][1] float: 'value of residual of range of current mission study loop'
            *  [7][0] string: 'residual_fl_initial_cruise_study',        [7][1] float: 'value of residual of cruise flight level of current mission study loop'
            *  [8][0] string: 'status_residual_m_take_off_max',          [8][1] bool:  'status of convergence for study take of mass'
            *  [9][0] string: 'status_residual_mission_energy',          [9][1] bool:  'status of convergence for study mission energy'
            * [10][0] string: 'status_range_study',                     [10][1] bool:  'status of convergence for mission study range'
            * [11][0] string: 'status_fl_initial_cruise_study',         [11][1] bool:  'status of convergence of initial mission study flight level'

        The input int "count_of_mission_study_loop" contains the current number of calibration loop.

        :param: paths_and_names: input list of lists
        :param: mission_study_variables: input list of lists
        :param: count_of_mission_study_loop: input int
        :return: none
    """

    ''' imports for python '''
    import os

    ''' read data for script execution '''
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]
    path_of_working_directory_rce = \
        ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]

    ''' initialize local parameter '''
    results = []
    reporting_of_ome_calibration = path_of_working_directory_rce + current_workflow_name \
        + '/temporaryResults/reporting/'

    ''' check if output temporary reporting folder exist '''
    if not os.path.isdir(reporting_of_ome_calibration):
        os.makedirs(reporting_of_ome_calibration)

    ''' prepare output data for csv-file '''
    # check if first loop of iteration -> if true: -> generate file header
    if count_of_mission_study_loop == 0:
        # generate header
        results.append('count_of_iteration; '
                       'm_take_off_max in kg; '
                       'residual_m_take_off_max; '
                       'mission_energy in J; '
                       'residual_mission_energy; '
                       'fl_initial_cruise; '
                       'residual_fl_initial_cruise; '
                       'range in nm; '
                       'residual_range;')

        # save initial values to csv-file
        results.append(str(count_of_mission_study_loop) + ';' +
                       str(0.00) +
                       'nan;' +
                       str(0.00) +
                       'nan;' +
                       str(0.00) + ';' +
                       'nan;' +
                       str(0.000) + ';' +
                       'nan;')

    count_of_mission_study_loop += 1

    results.append(str(count_of_mission_study_loop) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'm_take_off_max_study' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'residual_m_take_off_max_study' in item])[-1][-1], 6)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'mission_energy_study' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'residual_mission_energy_study' in item])[-1][-1], 6)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'fl_initial_cruise_study' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'residual_fl_initial_cruise_study' in item])[-1][-1], 1)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'range_study' in item])[-1][-1], 2)) + ';' +
                   str(round(([item for item in mission_study_variables
                               if 'residual_range_study' in item])[-1][-1], 6)) + ';')

    results_file = open(reporting_of_ome_calibration + 'study_loop_results.csv', 'a')
    for row in results:
        results_file.write(row + '\n')
    results_file.close()

    return
