# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def prepare_resume_of_parameter_study(paths_and_names, path_to_folder_old_reference, temporary_working_directory,
                                      log_file_list):
    """ Function to prepare resuming an old parameter study from existing old parameter study output folder.

        The input list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The input string "path_to_folder_old_reference" contains the absolute path to the reference aircraft from old parameter study.

        The input string "temporary_working_directory" contains the absolute path to the temporary working directory of current workflow execution.

        The input list "log_file_list" contains all system prints of function: set_configuration_parameter.py.

        The output list of lists "paths_and_names" contains the following values:
            *  [0][0] string: 'path_to_python_scripts',          [0][1] string: 'the system path to python scripts'
            *  [1][0] string: 'path_of_working_directory',       [1][1] string: 'the system path of working directory'
            *  [2][0] string: 'path_of_working_directory_rce',   [2][1] string: 'the system path of working directory for rce'
            *  [3][0] string: 'path_of_origin_tool_directory',   [3][1] string: 'the system path to the root directory of UNICADO repositories'
            *  [4][0] string: 'current_workflow_name',           [4][1] string: 'workflow name of the current workflow execution'
            *  [5][0] string: 'path_to_aircraft_projects',       [5][1] string: 'the system path to the repository of aircraft design projects'
            *  [6][0] string: 'path_to_aircraft_design_tools',   [6][1] string: 'the system path to the repository of aircraft design tools'
            *  [7][0] string: 'path_to_software_tools',          [7][1] string: 'the system path to the repository of software tools'
            *  [8][0] string: 'path_to_json_files',              [8][1] string: 'the path to module json files in .rce system folder'
            *  [9][0] string: 'aircraft_exchange_directory',     [9][1] string: 'name of the directory, witch contains the current aircraft project'
            * [10][0] string: 'aircraft_exchange_file',         [10][1] string: 'name of the current used aircraft exchange file'
            * [11][0] string: 'aircraft_project',               [11][1] string: 'name of the aircraft project, which ist currently designed'
            * [12][0] string: 'path_to_project',                [12][1] string: 'the system path to current project folder'
            * [13][0] string: 'aircraft_type',                  [13][1] string: 'name of aircraft type'

        The output bool "abort_parameter_study_flag" contains the status to abort the parameter study or not.

        The output list "log_file_list" contains all system prints of functions: set_configuration_parameter.py and initialize_parameter_study.py.

        :param: paths_and_names: input list of lists
        :param: path_to_folder_old_reference: input string
        :param: temporary_working_directory: input string
        :param: log_file_list: input list
        :return: paths_and_names: input list of lists, abort_parameter_study_flag: output bool, log_file_list: output list
    """

    ''' imports for python '''
    import os
    import shutil
    from datetime import datetime
    from inspect import currentframe, getframeinfo
    from sub_function.path_check import path_check
    from sub_function.read_xml_file import read_xml_file

    ''' read data for script execution '''
    current_workflow_name = ([item for item in paths_and_names if 'current_workflow_name' in item])[-1][-1]
    path_of_working_directory_rce = \
        ([item for item in paths_and_names if 'path_of_working_directory_rce' in item])[-1][-1]

    ''' initialize local parameter '''
    root_of_workflow_tree = []
    abort_parameter_study_flag = False
    function_name = getframeinfo(currentframe()).function

    ''' read aircraft project name from old workflow configuration file '''
    if os.path.isfile(path_to_folder_old_reference + 'unicado_workflow_conf_start.xml'):
        frame_info = getframeinfo(currentframe())
        path = path_to_folder_old_reference + 'unicado_workflow_conf_start.xml'
        root_of_workflow_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                        'UNICADOworkflow_start', function_name,
                                                        frame_info.lineno, log_file_list,
                                                        current_workflow_name)

    elif os.path.isfile(path_to_folder_old_reference + 'unicado_workflow_conf.xml'):
        frame_info = getframeinfo(currentframe())
        path = path_to_folder_old_reference + 'unicado_workflow_conf.xml'
        root_of_workflow_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                        'unicado_workflow_conf', function_name,
                                                        frame_info.lineno, log_file_list,
                                                        current_workflow_name)

    else:
        abort_parameter_study_flag = True

    ''' copy reference files from old parameter study directory to current working directory '''
    if not abort_parameter_study_flag:
        # delete existing aircraft project files from temporary working directory of current workflow execution
        list_of_temporary_files = os.listdir(temporary_working_directory)
        for list_element in list_of_temporary_files:
            # check if current list element is a directory -> if true: -> delete directory and sub folder if exist
            if os.path.isdir(temporary_working_directory + '/' + list_element):
                shutil.rmtree(temporary_working_directory + '/' + list_element)

            # else condition: current list element is a file -> delete file
            else:
                # check if the current file is not the workflow log file -> if true: -> delete file
                if not list_element == 'unicado_workflow.log':
                    os.remove(temporary_working_directory + '/' + list_element)

        # read name of aircraft project from old parameter study
        aircraft_project = root_of_workflow_tree.find("./control_settings/aircraft_exchange_file_name/value").text
        ([item for item in paths_and_names if 'aircraft_project' in item])[-1][-1] = aircraft_project[:-4]

        # create new aircraft project folder in temporary working directory
        if not os.path.isdir(temporary_working_directory + '/' + aircraft_project[:-4]):
            os.makedirs(temporary_working_directory + '/' + aircraft_project[:-4] + '/clean_sheet_design')

        # create new temporary result folder in temporary working directory
        if not os.path.isdir(temporary_working_directory + '/temporaryResults'):
            os.mkdir(temporary_working_directory + '/temporaryResults')

        # create new temporary folder in temporary working directory
        if not os.path.isdir(temporary_working_directory + '/temp'):
            os.mkdir(temporary_working_directory + '/temp')

        # read elements from old parameter study directory
        list_of_elements = os.listdir(path_to_folder_old_reference)
        if not 'parameter_study_conf.xml' in list_of_elements:
            abort_parameter_study_flag = True

        for element in list_of_elements:
            # check if current element is a directory -> if true: -> perform copy of directory entries
            if os.path.isdir(path_to_folder_old_reference + element):
                # reset all configuration files to files of old parameter study
                if element == 'config_files':
                    if os.path.isdir(path_to_folder_old_reference + element + '/preExecution'):
                        list_of_cfg_folders = os.listdir(path_to_folder_old_reference + element + '/preExecution')
                        for sub_folder in list_of_cfg_folders:
                            list_of_cfg_files = os.listdir(path_to_folder_old_reference + element + '/preExecution/'
                                                           + sub_folder)
                            for cfg_file in list_of_cfg_files:
                                shutil.copyfile(path_to_folder_old_reference + element + '/preExecution/' + sub_folder +
                                                '/' + cfg_file, temporary_working_directory + '/' + cfg_file)

                                print(cfg_file + ' has been successfully restored from the old parameter study folder!')
                                log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                                                     + ': ' + cfg_file
                                                     + ' has been successfully restored from the old parameter '
                                                       'study folder!')

                                # set path of temporary working directory of current workflow execution to cfg-file
                                if not cfg_file == 'liftingLine_conf.xml':
                                    frame_info = getframeinfo(currentframe())
                                    path = temporary_working_directory + '/' + cfg_file
                                    root_of_cfg_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                                               cfg_file[:-4], function_name,
                                                                               frame_info.lineno, log_file_list,
                                                                               current_workflow_name)

                                    root_of_cfg_tree.find(
                                        './control_settings/aircraft_exchange_file_directory/value').text = \
                                        temporary_working_directory + '/' + aircraft_project[:-4] + '/'
                                    xml_tree.write(path, encoding='utf-8')

                    else:
                        abort_parameter_study_flag = True

                # reset engine files to files of old parameter study
                if element == 'engine_data':
                    shutil.copytree(path_to_folder_old_reference + element,
                                    temporary_working_directory + '/' + aircraft_project[:-4] + '/engine_data')

                else:
                    abort_parameter_study_flag = True

                # reset all lean sheet design files to files of old parameter study
                if element == 'results':
                    list_of_results = os.listdir(path_to_folder_old_reference + element)
                    for result in list_of_results:
                        if os.path.isdir(path_to_folder_old_reference + element + '/' + result):
                            shutil.copytree(path_to_folder_old_reference + element + '/' + result,
                                            temporary_working_directory + '/' + aircraft_project[:-4]
                                            + '/clean_sheet_design/' + result)
                        else:
                            shutil.copyfile(path_to_folder_old_reference + element + '/' + result,
                                            temporary_working_directory + '/' + aircraft_project[:-4]
                                            + '/clean_sheet_design/' + result)

            # else condition: current element is a file -> copy file to temporary working directory
            else:
                if not element == 'unicado_workflow_conf_start.xml':
                    # copy project xml-files to temporary working directory
                    if aircraft_project[:-4] in element:
                        # check if the current element contains the string 'CPACS'
                        #  -> if true: -> copy cpacs file to project folder
                        if 'CPACS' in element:
                            shutil.copyfile(path_to_folder_old_reference + element,
                                            temporary_working_directory + '/' + aircraft_project[:-4] + '/' + element)

                        else:
                            abort_parameter_study_flag = True

                        # check if the current element contains the string 'start'
                        #  -> if true: -> copy aircraft exchange file to project folder
                        if 'start' in element:
                            # copy start version and rename it to working version of aircraft exchange file
                            shutil.copyfile(path_to_folder_old_reference + element,
                                            temporary_working_directory + '/' + aircraft_project[:-4] + '/'
                                            + aircraft_project)

                            # copy start version of aircraft exchange file
                            shutil.copyfile(path_to_folder_old_reference + element,
                                            temporary_working_directory + '/' + aircraft_project[:-4] + '/'
                                            + aircraft_project[:-4] + '_start.xml')
                        else:
                            abort_parameter_study_flag = True

                    if element == 'parameter_study_conf.xml':
                        frame_info = getframeinfo(currentframe())
                        path = path_to_folder_old_reference + element
                        root_of_parameter_study_tree, xml_tree = read_xml_file(path, path_of_working_directory_rce,
                                                                               'parameter_study', function_name,
                                                                               frame_info.lineno, log_file_list,
                                                                               current_workflow_name)

                        number_of_input_parameter = int(root_of_parameter_study_tree.find(
                            "./ProgramSettings/ParameterSettings/InputParameters").text)
                        number_of_output_parameter = int(root_of_parameter_study_tree.find(
                            "./ProgramSettings/ParameterSettings/OutputParameters").text)
                        path_to_current_study = str(root_of_parameter_study_tree.find(
                            "./ProgramSettings/SensitivityStudySettings/PathToCurrentStudy").text)
                        _, path_to_current_study = path_check(path_to_current_study)
                        root_of_parameter_study_tree.find(
                            "./ProgramSettings/SensitivityStudySettings/ReproduceOutputFileFromOldStudy").text = str(1)
                        root_of_parameter_study_tree.find(
                            "./ProgramSettings/SensitivityStudySettings/PathToFolderOldStudy").text \
                            = path_to_current_study

                        # loop across all input parameter to set paths
                        i = 1
                        while i <= number_of_input_parameter:
                            path_in_xml_file = str(root_of_parameter_study_tree.find(
                                "./ProgramSettings/ParameterSettings/InputParameter[" + "@ID='" + str(i) + "'"
                                + "]PathInXmlFile").text)

                            # check if the string "ConfigFile" in the given xml-path
                            #  -> if true: -> set path to aircraft exchange file
                            if "AcftExchangeFile" in path_in_xml_file:
                                root_of_parameter_study_tree.find(
                                    "./ProgramSettings/ParameterSettings/InputParameter[" + "@ID='"
                                    + str(i) + "'" + "]RelDirectory").text \
                                    = str('../workingDirectoryRCE/' + current_workflow_name + '/'
                                          + aircraft_project[:-4])

                            # check if the string "ConfigFile" in the given xml-path
                            #  -> if true: -> set path to module configuration file
                            elif "ConfigFile" in path_in_xml_file:
                                root_of_parameter_study_tree.find(
                                    "./ProgramSettings/ParameterSettings/InputParameter[" + "@ID='"
                                    + str(i) + "'" + "]RelDirectory").text \
                                    = str('../workingDirectoryRCE/' + current_workflow_name)
                            i += 1

                        # loop across all output parameter to set paths
                        i = 1
                        while i <= number_of_output_parameter:
                            path_in_xml_file = str(root_of_parameter_study_tree.find(
                                "./ProgramSettings/ParameterSettings/OutputParameter[" + "@ID='" + str(i) + "'"
                                + "]PathInXmlFile").text)

                            # check if the string "ConfigFile" in the given xml-path
                            #  -> if true: -> set path to aircraft exchange file
                            if "AcftExchangeFile" in path_in_xml_file:
                                root_of_parameter_study_tree.find(
                                    "./ProgramSettings/ParameterSettings/OutputParameter[" + "@ID='"
                                    + str(i) + "'" + "]RelDirectory").text \
                                    = str('../workingDirectoryRCE/' + current_workflow_name + '/'
                                          + aircraft_project[:-4])

                            # check if the string "ConfigFile" in the given xml-path
                            #  -> if true: -> set path to module configuration file
                            elif "ConfigFile" in path_in_xml_file:
                                root_of_parameter_study_tree.find(
                                    "./ProgramSettings/ParameterSettings/OutputParameter[" + "@ID='"
                                    + str(i) + "'" + "]RelDirectory").text \
                                    = str('../workingDirectoryRCE/' + current_workflow_name)
                            i += 1

                        xml_tree.write(path, encoding='utf-8')

                    # copy other necessary files from old study to temporary working directory
                    shutil.copyfile(path_to_folder_old_reference + element,
                                    temporary_working_directory + '/' + element)

    return paths_and_names, abort_parameter_study_flag, log_file_list
