# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def set_input_parameter_settings(install_path_directory, path_to_current_study, input_parameter_settings,
                                 xml_path_status, log_file_list, current_workflow_name='UNICADOworkflow'):
    """ Function to set study parameter to necessary xml-file.

        The input string 'install_path_directory' contains the absolute path to the UNICADO install directory.

        The input string 'path_to_current_study' contains the absolute path to the working directory of current parameter study.

        The input list of lists "input_parameter_settings" contains ato following values:
            *  [0][0] string: 'parameter_name',                  [0][1] string: 'name of the current study input parameter'
            *  [1][0] string: 'input_directory_name',            [1][1] string: 'name of the input directory of current study parameter'
            *  [2][0] string: 'input_file_name'                  [2][1] string: 'name of the input file of the current study parameter'
            *  [3][0] string: 'path_in_xml_file'                 [3][1] string: 'path to the study parameter in the input xml-file'
            *  [4][0] string: 'step_type'                        [4][1] string: 'mode for step value handling'
            *  [5][0] string: 'read_values_from_text_file',      [5][1] int:    'switch to read iteration values from an external data file'
            *  [6][0] string: 'step_size',                       [8][1] float:  'value of step size of current study parameter'
            *  [7][0] string: 'number_of_steps_up',              [9][1] int:    'value of maximum iteration steps upwards are to be performed'
            *  [8][0] string: 'number_of_steps_down',           [10][1] int:    'value of maximum iteration steps downwards are to be performed'
            *  [9][0] string: 'study_counter'                   [11][1] int:    'number of iteration loop of current input parameter'

        The input bool 'xml_path_status' contains the status "True" of current xml-path.

        The input list "log_file_list" contains all system prints of function: perform_parameter_study.py.

        The input string "current_workflow_name" contains the name of the current workflow execution.
            * If no input is given, default string 'UNICADOworkflow' is used.

        The output bool 'xml_path_status' contains the current status "True" or "False" of current xml-path.

        The output float 'new_value_of_input' contains the new parameter iteration value of current loop.

        The output list "log_file_list" contains all system prints of functions: perform_parameter_study.py and read_parameter_study_settings.py.

        :param: install_path_directory: input string
        :param: path_to_current_study: input string
        :param: input_parameter_settings: input list of lists
        :param: xml_path_status: input bool
        :param: log_file_list: input list
        :param: current_workflow_name: input string
        :return: xml_path_status: output bool, new_value_of_input: output float, log_file_list: output list
    """

    ''' imports for python '''
    from inspect import currentframe, getframeinfo
    from datetime import datetime
    from sub_function.path_check import path_check
    from sub_function.read_xml_file import read_xml_file
    from sub_function.xml_path_check import xml_path_check
    from parameter_study.estimate_input_value_from_configuration_file \
        import estimate_input_value_from_configuration_file

    ''' read data for script execution '''
    input_file_name = ([item for item in input_parameter_settings if 'input_file_name' in item])[-1][-1]
    path_in_xml_file = ([item for item in input_parameter_settings if 'path_in_xml_file' in item])[-1][-1]
    input_directory_name = ([item for item in input_parameter_settings if 'input_directory_name' in item])[-1][-1]
    parameter_name = ([item for item in input_parameter_settings if 'parameter_name' in item])[-1][-1]

    ''' initialize local parameter '''
    new_value_of_input = float(0.0)
    number_of_runs_to_perform = 1
    function_name = getframeinfo(currentframe()).function
    path_of_working_directory_rce = install_path_directory + 'workingDirectoryRCE/'

    # call function to check if the given path in relative style otherwise correct the path
    corrected_path_is_absolute, corrected_path = path_check(input_directory_name)

    # call function to check path in xml-file and correct to unix path
    corrected_xml_path = xml_path_check(path_in_xml_file)

    # set absolute path of study xml-file
    #if corrected_path_is_absolute:
    #    path_to_study_xml_file = corrected_path + input_file_name
    #else:
    #    path_to_study_xml_file = path_of_working_directory_rce + current_workflow_name + input_file_name
    if "projects" in corrected_path:
        corrected_path = corrected_path[corrected_path.rfind('/', 0, -1):]
    else:
        corrected_path = '/'
    path_to_study_xml_file = path_of_working_directory_rce + current_workflow_name + corrected_path + input_file_name
    print("Input variable path: " + path_to_study_xml_file)

    # read study input xml-file as element tree
    frame_info = getframeinfo(currentframe())
    root_of_study_xml_file, xml_tree = read_xml_file(path_to_study_xml_file, path_of_working_directory_rce,
                                                     input_file_name, function_name, frame_info.lineno, log_file_list,
                                                     current_workflow_name)

    ''' set iteration value to study parameter xml-file '''
    # check if the corrected xml-path is inside the current study element tree
    #  -> if true: -> set study parameter in xml-file
    if not root_of_study_xml_file.find(corrected_xml_path) is None:
        # call function to set input parameter from parameter study configuration file
        new_value_of_input, xml_path_status, number_of_runs_to_perform, log_file_list = \
            estimate_input_value_from_configuration_file(root_of_study_xml_file, input_parameter_settings,
                                                         path_to_current_study, corrected_xml_path, log_file_list)

        # check if the estimation of new input value failed
        #  -> if true: -> raise an error and skipp current input parameter
        if not xml_path_status:
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': '
                                 'Error: No steps up and no steps down are set in the parameter study '
                                 'configuration file! \n                           '
                                 '       Current study input parameter will be skipped!')
            print('Error: No steps up and no steps down for parameter "' + parameter_name +
                  '" are set in the parameter study configuration file! \n'
                  '       Current study input parameter will be skipped!')

        # else condition: new value of current input parameter could be estimated
        #  -> write value to current study xml-file
        else:
            print('New value of current parameter study input is : ' + str(new_value_of_input))
            log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': '
                                 'New value of current parameter study input is: ' + str(new_value_of_input))
            xml_tree.find(corrected_xml_path).text = str(new_value_of_input)
            xml_tree.write(path_to_study_xml_file, encoding='utf-8')

    # else condition: corrected xml-path is not inside the current study element tree
    #  -> raise an error and skip current parameter study
    else:
        xml_path_status = False
        log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': '
                             'Error: The given xml-path in the parameter study configuration file of current input '
                             'parameter is not existing! \n                           '
                             '       Current study input parameter will be skipped!')
        print('Error: The given xml-path in the parameter study configuration file of current input parameter '
              'is not existing! \n'
              '       Current study input parameter will be skipped!')

    return xml_path_status, new_value_of_input, number_of_runs_to_perform, log_file_list
