# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def read_xml_file(path, path_of_working_directory_rce, xml_file_name, function_name, code_line, log_file_list,
                  current_workflow_name='UNICADOworkflow'):
    """ function to read data from given xml-file as element tree.

        The input string "path" contains the absolute path to the given xml-file.

        The input string "path_of_working_directory_rce" contains the absolute path to the working directory of rce.

        The input string "xml_file_name" contains the name of the given xml-file to read.

        The input string "function_name" contains the name of the python function, witch has called the read_xml_file.py.

        The input integer "code_line" contains the code line number of the python function, witch has called the read_xml_file.py.

        The input list "log_file_list" contains the strings of workflow log-file from the caller function and the added strings from this function.

        The input string "current_workflow_name" contains the name of the current workflow execution.
            * If no input is given, default sting 'UNICADOworkflow' is used.

        The output elementTree "root_of_tree" contains the data of the given xml-file as an elementTree.

        The output elementTree "xml_tree" contains the data of the given xml-file as an elementTree.

        :param: path: input string
        :param: path_of_working_directory_rce: input string
        :param: xml_file_name: input string
        :param: function_name: input string
        :param: code_line: input integer
        :param: log_file_list: input list
        :param: current_workflow_name: input string
        :return: root_of_tree: output elementTree, xml_tree: output elementTree
    """

    """ imports for python """
    import sys
    import xml.etree.ElementTree
    from datetime import datetime
    from sub_function.write_to_log_file import write_to_log_file

    ''' read xml-file as element tree '''
    try:
        xml_tree = xml.etree.ElementTree.ElementTree(file=path)
        root_of_tree = xml_tree.getroot()

    except OSError:
        print('Error in file: ' + function_name + '.py \n In line: ' + str(code_line + 2) + ' \n ' + xml_file_name
              + ' configuration file could not be opened!  \n Program aborted!')
        log_file_list.append(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ': Error in file: ' + function_name
                             + '.py \n In line: ' + str(code_line + 2) + ' \n ' + xml_file_name
                             + ' configuration file could not be opened!  \n Program aborted!')
        write_to_log_file(path_of_working_directory_rce, current_workflow_name, log_file_list)
        sys.exit(1)

    return root_of_tree, xml_tree
